/***************************************************************************
 *   Copyright (C) 2008, 2009 by Stephane List                             *
 *   slist@lilotux.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "qmotion.h"

#include <ctime>

#include <QMessageBox>
#include <QSettings>
#include <QDebug>
#include <QDir>
#include <QDate>
#include <QTime>
#include <QColorDialog>
#include <QPainter>
#include <QDateTime>

// various tracking parameters (in seconds)
const double MHI_DURATION = 1;
const double MAX_TIME_DELTA = 0.5;
const double MIN_TIME_DELTA = 0.05;
// number of cyclic frame buffer used for motion detection
// (should, probably, depend on FPS)
const int N = 4;
const int MAGNITUDE_GLOBAL = 100;
const int MAGNITUDE_COMPONENT = 30;
const QString QMotion_version = "1.9";


QMotion::QMotion(QWidget *parent)
		: QMainWindow(parent),
		image_(0),
		motion_(0),
		counter_(0),
		counter_last_(0),
		buf(0),
		last(0),
		mhi(0),
		orient(0),
		mask(0),
		red(0),
		green(0),
		blue(0),
		segmask(0),
		storage(0)
{
	QSettings settings;
	setupUi(this);

	color_ = QColor(settings.value("color", Qt::blue).toString());
	frame_color->setPalette(QPalette(color_));
	frame_color->setAutoFillBackground(true);
	
	limit_fps->setValue(settings.value("limit_fps", 25).toInt());
	checkBox_timestamp->setChecked(settings.value("timestamp", false).toBool());
	Flip->setChecked(settings.value("flip",0).toBool());
	threshold->setValue(settings.value("detection_threshold", 30).toInt());

	connect(actionQuit, SIGNAL(triggered()), qApp, SLOT(quit()));
	connect(actionAbout_QT, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
	connect(actionAbout_QMotion, SIGNAL(triggered()), this, SLOT(about()));
	connect(action_Directory_settings, SIGNAL(triggered()), this, SLOT(dir_settings()));

	connect(this, SIGNAL(motion()), this, SLOT(motion_treatment()));

	capture_ = cvCaptureFromCAM(0);
	if (!capture_)
	{
    qWarning("No webcam");
		label_video->setText(tr("No webcam"));
		label_video->setAlignment(Qt::AlignCenter);
		label_motion->setText("");
		return;
	}
	else
	{
		label_video->setText(tr("Webcam OK!"));
	}
	cvGrabFrame(capture_);

	// get image to have size
	image_ = cvRetrieveFrame(capture_);
	label_size->setText(tr("size: %1%2").arg(image_->width).arg(image_->height));

//	label_video->setMinimumHeight(image_->height * image_->width / label_video->width());

	timer_.start(1000 / limit_fps->value());
	timer_fps_.start(1000);

	connect(&timer_,SIGNAL(timeout()), this,SLOT(captureLoop()));
	connect(&timer_fps_,SIGNAL(timeout()), this,SLOT(fps_update()));
}

QMotion::~QMotion()
{
	QSettings settings;
	settings.setValue("limit_fps", limit_fps->value());
	settings.setValue("timestamp", checkBox_timestamp->isChecked());
	settings.setValue("color", color_.name());
	settings.setValue("flip",Flip->isChecked()); 
	settings.setValue("detection_threshold",threshold->value());

	if (capture_)
	{
		cvReleaseCapture(&capture_);
	}
}


QImage QMotion::Ipl2QImage(const IplImage* newImage,bool upsideDown)
{
  if ((0 != newImage) && (cvGetSize(newImage).width > 0))
  {
    const char* src = newImage->imageData;
		QImage qtemp(newImage->width,newImage->height,QImage::Format_RGB32);

		if (upsideDown)
		{
			for (int y = 0; y < newImage->height; ++y, src += newImage->widthStep)
			{
				uint* dst = reinterpret_cast<uint*>(qtemp.scanLine (newImage->height - y - 1));
				for (int x = 0; x < newImage->width; ++x, ++dst)
				{
					*dst = qRgb(src[x * newImage->nChannels + 2]
							,src[x * newImage->nChannels + 1]
							,src[x * newImage->nChannels]);
				}
			}
		}
		else
		{
			for (int y = 0; y < newImage->height; ++y, src += newImage->widthStep)
			{
				uint* dst = reinterpret_cast<uint*>(qtemp.scanLine(y));
				for (int x = 0; x < newImage->width; ++x, ++dst)
				{
					*dst = qRgb(src[x * newImage->nChannels + 2]
							,src[x * newImage->nChannels + 1]
							,src[x * newImage->nChannels]);
				}
			}
		}
		return qtemp;
  }
	else
		return QImage();
}

void QMotion::add_timestamp(QImage & img)
{
	QFont def_font;
	QFontMetrics fm(def_font);
	int textHeightInPixels = fm.height();

	QPainter painter(&img);
	QRect bottomrect(0, img.rect().bottom() - textHeightInPixels, img.width(), textHeightInPixels);
	painter.setOpacity(0.5);
	painter.fillRect(bottomrect, Qt::white);

	painter.setOpacity(1);
	painter.setPen(Qt::black);

	QDate d(QDate::currentDate());
	QTime t(QTime::currentTime());
	painter.drawText(img.rect(), Qt::AlignHCenter|Qt::AlignBottom, d.toString(Qt::DefaultLocaleLongDate) + " " + t.toString("hh:mm:ss.zzz"));
}

void QMotion::captureLoop()
{
	if ( !capture_ )
	{
		return;
	}
	cvGrabFrame(capture_);
	image_ = cvRetrieveFrame(capture_);
	if (image_)
	{
		counter_ ++;
		if (!checkBox_disable_display->isChecked())
		{
			qImage_ = Ipl2QImage(image_,Flip->isChecked());
			if (checkBox_timestamp->isChecked())
			{
				add_timestamp(qImage_);
			}
			label_video->setPixmap(QPixmap::fromImage(qImage_.scaled(label_video->width(),label_video->height()),Qt::AutoColor));
		}
		if( !motion_ )
		{
			motion_ = cvCreateImage(cvSize(image_->width,image_->height), 8, 3 );
			cvZero(motion_);
			motion_->origin = image_->origin;
		}
		//update_mhi( image_, motion_, 30 );
		update_mhi( image_, motion_, threshold->value() );
		if (!checkBox_disable_display->isChecked())
		{
			qMotion_ = Ipl2QImage(motion_,Flip->isChecked());
			label_motion->setPixmap(QPixmap::fromImage(qMotion_.scaled(label_motion->width(),label_motion->height()),Qt::AutoColor));
		}
	}
}

void QMotion::fps_update()
{
	lcdNumber->display(counter_ - counter_last_);
	counter_last_ = counter_;
}

// parameters:
//  img - input video frame
//  dst - resultant motion picture
//  args - optional parameters
void QMotion::update_mhi( IplImage* img, IplImage* dst, int diff_threshold )
{
    double timestamp = (double)clock()/CLOCKS_PER_SEC; // get current time in seconds
    CvSize size = cvSize(img->width,img->height); // get current frame size
    int i, idx1 = last, idx2;
    IplImage* silh;
    CvSeq* seq;
    CvRect comp_rect;
    double count;
    double angle;
    CvPoint center;
    double magnitude;
    CvScalar color;

    // allocate images at the beginning or
    // reallocate them if the frame size is changed
    if( !mhi || mhi->width != size.width || mhi->height != size.height ) {
        if( buf == 0 ) {
            buf = (IplImage**)malloc(N*sizeof(buf[0]));
            memset( buf, 0, N*sizeof(buf[0]));
        }

        for( i = 0; i < N; i++ ) {
            cvReleaseImage( &buf[i] );
            buf[i] = cvCreateImage( size, IPL_DEPTH_8U, 1 );
            cvZero( buf[i] );
        }
        cvReleaseImage( &mhi );
        cvReleaseImage( &orient );
        cvReleaseImage( &segmask );
        cvReleaseImage( &mask );
        cvReleaseImage( &red );
        cvReleaseImage( &green );
        cvReleaseImage( &blue );

        mhi = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        cvZero( mhi ); // clear MHI at the beginning
        orient = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        segmask = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        mask = cvCreateImage( size, IPL_DEPTH_8U, 1 );
        red = cvCreateImage( size, IPL_DEPTH_8U, 1 );
        green = cvCreateImage( size, IPL_DEPTH_8U, 1 );
        blue = cvCreateImage( size, IPL_DEPTH_8U, 1 );
    }


    cvCvtColor( img, buf[last], CV_BGR2GRAY ); // convert frame to grayscale

    idx2 = (last + 1) % N; // index of (last - (N-1))th frame
    last = idx2;

    silh = buf[idx2];
    cvAbsDiff( buf[idx1], buf[idx2], silh ); // get difference between frames

    cvThreshold( silh, silh, diff_threshold, 1, CV_THRESH_BINARY ); // and threshold it
    cvUpdateMotionHistory( silh, mhi, timestamp, MHI_DURATION ); // update MHI

    // convert MHI to blue 8u image
    cvCvtScale( mhi, mask, 255./MHI_DURATION, (MHI_DURATION - timestamp)*255./MHI_DURATION );

		// convert 8u image to 32u image using custom color
    cvCvtScale( mask, red,   color_.red() / 255.0, 0.);
    cvCvtScale( mask, green, color_.green() / 255.0, 0.);
    cvCvtScale( mask, blue,  color_.blue() / 255.0, 0.);

    cvZero( dst );
    cvCvtPlaneToPix( blue, green,  red, 0, dst );

    // calculate motion gradient orientation and valid orientation mask
    cvCalcMotionGradient( mhi, mask, orient, MAX_TIME_DELTA, MIN_TIME_DELTA, 3 );

    if( !storage )
        storage = cvCreateMemStorage(0);
    else
        cvClearMemStorage(storage);

    // segment motion: get sequence of motion components
    // segmask is marked motion components map. It is not used further
    seq = cvSegmentMotion( mhi, segmask, storage, timestamp, MAX_TIME_DELTA );

    // iterate through the motion components,
    // One more iteration (i == -1) corresponds to the whole image (global motion)
    for( i = -1; i < seq->total; i++ ) {

        if( i < 0 ) { // case of the whole image
            comp_rect = cvRect( 0, 0, size.width, size.height );
            color = CV_RGB(255,255,255); // white
            magnitude = MAGNITUDE_GLOBAL;
        }
        else { // i-th motion component
            comp_rect = ((CvConnectedComp*)cvGetSeqElem( seq, i ))->rect;
            if( comp_rect.width + comp_rect.height < 100 ) // reject very small components
                continue;
            color = CV_RGB(255,0,0); // red
            magnitude = MAGNITUDE_COMPONENT;
        }

        // select component ROI
        cvSetImageROI( silh, comp_rect );
        cvSetImageROI( mhi, comp_rect );
        cvSetImageROI( orient, comp_rect );
        cvSetImageROI( mask, comp_rect );

        // calculate orientation
        angle = cvCalcGlobalOrientation( orient, mask, mhi, timestamp, MHI_DURATION);
        angle = 360.0 - angle;  // adjust for images with top-left origin

        count = cvNorm( silh, 0, CV_L1, 0 ); // calculate number of points within silhouette ROI

        cvResetImageROI( mhi );
        cvResetImageROI( orient );
        cvResetImageROI( mask );
        cvResetImageROI( silh );

        // check for the case of little motion
        if( count < comp_rect.width*comp_rect.height * 0.05 )
            continue;

				if (! checkBox_disable_display->isChecked()
					&& (((magnitude == MAGNITUDE_GLOBAL) && global_marker->isChecked())
						|| ((magnitude == MAGNITUDE_COMPONENT) && component_markers->isChecked())))
				{
					// draw a clock with arrow indicating the direction
					center = cvPoint(comp_rect.x + comp_rect.width / 2,comp_rect.y + comp_rect.height / 2);

					magnitude = (magnitude * size.height) / (3 * MAGNITUDE_GLOBAL);

					cvCircle(dst,center,cvRound(magnitude * 1.2),color,3,CV_AA,0);
					cvLine(dst,center
						,cvPoint(cvRound(center.x + magnitude * cos(angle * CV_PI / 180))
							,cvRound(center.y - magnitude * sin(angle * CV_PI / 180)))
						,color,3,CV_AA,0);
				}
				if (i == -1)
				{
					emit motion();
				}
    }
}

void QMotion::about()
{
	QMessageBox::about(this,
			tr("About"),
			tr("<p><b>QMotion</b></p>"
				"<p>Version QMotion: %1"
				"<p>Version OpenCV: %2"
				"<p>Version %3: %4.%5"
				"</p>")
			.arg(QMotion_version)
			.arg(CV_VERSION)
#ifdef __GNUC__
			.arg("GCC")
			.arg(__GNUC__)
			.arg(__GNUC_MINOR__)
#elif defined(_MSC_VER)
			.arg("VC++")
			.arg(_MSC_VER / 100)
			.arg(_MSC_VER % 100,2,10,QChar('0'))
#else
			.arg("Compiler")
			.arg("?")
			.arg("?")
#endif
			);
}

void QMotion::motion_treatment()
{
	if (checkBox_save->isChecked())
	{
		QSettings settings;
		QString snapshot_file;
		settings.beginGroup("recorder");
		snapshot_file = settings.value("dir", QDir::homePath()).toString();
		if (!snapshot_file.endsWith('/'))
		{
			snapshot_file += "/";
		}
		snapshot_file += "snapshot__" + QDate::currentDate().toString("yyyy_MM_dd") + "__" + QTime::currentTime().toString("hh_mm_ss_zzz") + ".png";
		settings.endGroup();
		if (checkBox_disable_display->isChecked())
		{
			qImage_ = Ipl2QImage(image_,Flip->isChecked());
			add_timestamp(qImage_);
		}
		if (qImage_.save(snapshot_file))
		{
			qDebug() << "Saved in " << snapshot_file;
		}
		else
		{
			qWarning() << "Error saving in " << snapshot_file;
		}
	}
}

void QMotion::dir_settings()
{
	RecorderSetupDlg dialog(0,  Qt::WindowTitleHint | Qt::WindowSystemMenuHint);
	dialog.exec();
}

void QMotion::on_pushButton_color_clicked()
{
	QColor color = QColorDialog::getColor(color_, this);
	if (color.isValid()) {
		frame_color->setPalette(QPalette(color));
		frame_color->setAutoFillBackground(true);
		color_ = color;
	}
}

void QMotion::on_limit_fps_valueChanged(int limite)
{
	timer_.setInterval(1000/limite);
}

void QMotion::resizeEvent( QResizeEvent * event )
{
  QWidget::resizeEvent(event);
  if ((image_) && (image_->width != 0))
  {
    label_video->setMinimumHeight(image_->height * label_video->width() / image_->width);
    label_video->setMaximumHeight(image_->height * label_video->width() / image_->width);
    label_motion->setMinimumHeight(image_->height * label_motion->width() / image_->width);
    label_motion->setMaximumHeight(image_->height * label_motion->width() / image_->width);
  }
}
